@@ -8,7 +8,7 @@
  * No Difference
  * 
  * The user file looks modified judging from its time stamp; however it needn't
- * be.  No_difference() finds out whether it is or not. If it is not, it
+ * be.  No_Difference() finds out whether it is or not. If it is not, it
  * updates the administration.
  * 
  * returns 0 if no differences are found and non-zero otherwise
@@ -49,8 +49,9 @@ No_Difference (finfo, vers)
 	options = xstrdup ("");
 
     tocvsPath = wrap_tocvs_process_file (finfo->file);
-    retcode = RCS_cmp_file (vers->srcfile, vers->vn_user, options,
-			    tocvsPath == NULL ? finfo->file : tocvsPath);
+    retcode = RCS_cmp_file( vers->srcfile, vers->vn_user, (char **)NULL,
+                            (char *)NULL, options,
+			    tocvsPath == NULL ? finfo->file : tocvsPath );
     if (retcode == 0)
     {
 	/* no difference was found, so fix the entries file */
@@ -70,7 +71,7 @@ No_Difference (finfo, vers)
 
 	/* update the entdata pointer in the vers_ts structure */
 	p = findnode (finfo->entries, finfo->file);
-	vers->entdata = (Entnode *) p->data;
+	vers->entdata = p->data;
 
 	ret = 0;
     }
