@@ -49,30 +49,29 @@ static int checkout_file PROTO ((struct file_info *finfo, Vers_TS *vers_ts,
 				 int adding, int merging, int update_server));
 #ifdef SERVER_SUPPORT
 static void checkout_to_buffer PROTO ((void *, const char *, size_t));
-#endif
-#ifdef SERVER_SUPPORT
 static int patch_file PROTO ((struct file_info *finfo,
 			      Vers_TS *vers_ts, 
 			      int *docheckout, struct stat *file_info,
 			      unsigned char *checksum));
 static void patch_file_write PROTO ((void *, const char *, size_t));
-#endif
+#endif /* SERVER_SUPPORT */
 static int merge_file PROTO ((struct file_info *finfo, Vers_TS *vers));
 static int scratch_file PROTO((struct file_info *finfo, Vers_TS *vers));
-static Dtype update_dirent_proc PROTO ((void *callerdat, char *dir,
-					char *repository, char *update_dir,
-					List *entries));
-static int update_dirleave_proc PROTO ((void *callerdat, char *dir,
-					int err, char *update_dir,
+static Dtype update_dirent_proc PROTO ((void *callerdat, const char *dir,
+                                        const char *repository,
+                                        const char *update_dir,
+                                        List *entries));
+static int update_dirleave_proc PROTO ((void *callerdat, const char *dir,
+					int err, const char *update_dir,
 					List *entries));
 static int update_fileproc PROTO ((void *callerdat, struct file_info *));
 static int update_filesdone_proc PROTO ((void *callerdat, int err,
-					 char *repository, char *update_dir,
-					 List *entries));
+                                         const char *repository,
+                                         const char *update_dir,
+                                         List *entries));
 #ifdef PRESERVE_PERMISSIONS_SUPPORT
 static int get_linkinfo_proc PROTO ((void *callerdat, struct file_info *));
 #endif
-static void write_letter PROTO ((struct file_info *finfo, int letter));
 static void join_file PROTO ((struct file_info *finfo, Vers_TS *vers_ts));
 
 static char *options = NULL;
@@ -184,7 +183,7 @@ update (argc, argv)
 #endif
 		    error (1, 0,
 			   "-q or -Q must be specified before \"%s\"",
-			   command_name);
+			   cvs_cmd_name);
 		break;
 	    case 'd':
 		update_build_dirs = 1;
@@ -291,6 +290,8 @@ update (argc, argv)
 		if (supported_request ("update-patches"))
 		    send_arg ("-u");
 
+		send_arg ("--");
+
                 if (update_build_dirs)
                     flags |= SEND_BUILD_DIRS;
 
@@ -318,6 +319,8 @@ update (argc, argv)
 		    error (1, errno, "could not chdir to %s", toplevel_wd);
 		}
 
+		send_arg ("--");
+
 		for (i = 0; i < failed_patches_count; i++)
 		    if (unlink_file (failed_patches[i]) < 0
 			&& !existence_error (errno))
@@ -410,22 +413,25 @@ update (argc, argv)
     /* call the command line interface */
     err = do_update (argc, argv, options, tag, date, force_tag_match,
 		     local, update_build_dirs, aflag, update_prune_dirs,
-		     pipeout, which, join_rev1, join_rev2, (char *) NULL, 1);
+		     pipeout, which, join_rev1, join_rev2, (char *) NULL, 1,
+		     (char *) NULL);
 
     /* free the space Make_Date allocated if necessary */
     if (date != NULL)
 	free (date);
 
-    return (err);
+    return err;
 }
 
+
+
 /*
  * Command line interface to update (used by checkout)
  */
 int
 do_update (argc, argv, xoptions, xtag, xdate, xforce, local, xbuild, xaflag,
 	   xprune, xpipeout, which, xjoin_rev1, xjoin_rev2, preload_update_dir,
-	   xdotemplate)
+	   xdotemplate, repository)
     int argc;
     char **argv;
     char *xoptions;
@@ -442,6 +448,7 @@ do_update (argc, argv, xoptions, xtag, xdate, xforce, local, xbuild, xaflag,
     char *xjoin_rev2;
     char *preload_update_dir;
     int xdotemplate;
+    char *repository;
 {
     int err = 0;
     char *cp;
@@ -488,10 +495,10 @@ do_update (argc, argv, xoptions, xtag, xdate, xforce, local, xbuild, xaflag,
 	   follows it; someone should make sure that I did it right. */
 	err = start_recursion (get_linkinfo_proc, (FILESDONEPROC) NULL,
 			       (DIRENTPROC) NULL, (DIRLEAVEPROC) NULL, NULL,
-			       argc, argv, local, which, aflag, 1,
-			       preload_update_dir, 1);
+			       argc, argv, local, which, aflag, CVS_LOCK_READ,
+			       preload_update_dir, 1, (char *) NULL);
 	if (err)
-	    return (err);
+	    return err;
 
 	/* FIXME-twp: at this point we should walk the hardlist
 	   and update the `links' field of each hardlink_info struct
@@ -504,8 +511,8 @@ do_update (argc, argv, xoptions, xtag, xdate, xforce, local, xbuild, xaflag,
     /* call the recursion processor */
     err = start_recursion (update_fileproc, update_filesdone_proc,
 			   update_dirent_proc, update_dirleave_proc, NULL,
-			   argc, argv, local, which, aflag, 1,
-			   preload_update_dir, 1);
+			   argc, argv, local, which, aflag, CVS_LOCK_READ,
+			   preload_update_dir, 1, repository);
 
 #ifdef SERVER_SUPPORT
     if (server_active)
@@ -518,7 +525,7 @@ do_update (argc, argv, xoptions, xtag, xdate, xforce, local, xbuild, xaflag,
 	sleep_past (last_register_time);
     }
 
-    return (err);
+    return err;
 }
 
 #ifdef PRESERVE_PERMISSIONS_SUPPORT
@@ -558,12 +565,14 @@ get_linkinfo_proc (callerdat, finfo)
     hlinfo->status = (Ctype) 0;	/* is this dumb? */
     hlinfo->checked_out = 0;
 
-    linkp->data = (char *) hlinfo;
+    linkp->data = hlinfo;
 
     return 0;
 }
 #endif
 
+
+
 /*
  * This is the callback proc for update.  It is called for each file in each
  * directory by the recursion code.  The current directory is the local
@@ -686,38 +695,8 @@ update_fileproc (callerdat, finfo)
                 {
                     if (vers->ts_conflict)
                     {
-                        char *filestamp;
-                        int retcode;
-
-                        /*
-                         * If the timestamp has changed and no
-                         * conflict indicators are found, it isn't a
-                         * 'C' any more.
-                         */
-
-#ifdef SERVER_SUPPORT
-                        if (server_active)
-                            retcode = vers->ts_conflict[0] != '=';
-                        else 
-                        {
-                            filestamp = time_stamp (finfo->file);
-                            retcode = strcmp (vers->ts_conflict, filestamp);
-                            free (filestamp);
-                        }
-#else
-                        filestamp = time_stamp (finfo->file);
-                        retcode = strcmp (vers->ts_conflict, filestamp);
-                        free (filestamp);
-#endif
-
-                        if (retcode)
-                        {
-                            /* The timestamps differ.  But if there
-                               are conflict markers print 'C' anyway.  */
-                            retcode = !file_has_markers (finfo);
-                        }
-
-                        if (!retcode)
+			if (file_has_conflict (finfo, vers->ts_conflict)
+			    || file_has_markers (finfo))
                         {
                             write_letter (finfo, 'C');
                             retval = 1;
@@ -732,10 +711,7 @@ update_fileproc (callerdat, finfo)
                         }
                     }
                     if (!retval)
-                    {
                         write_letter (finfo, 'M');
-                        retval = 0;
-                    }
                 }
 		break;
 	    case T_PATCH:		/* needs patch */
@@ -805,42 +781,48 @@ update_fileproc (callerdat, finfo)
     }
 
     freevers_ts (&vers);
-    return (retval);
+    return retval;
 }
 
-static void update_ignproc PROTO ((char *, char *));
+
+
+static void update_ignproc PROTO ((const char *, const char *));
 
 static void
 update_ignproc (file, dir)
-    char *file;
-    char *dir;
+    const char *file;
+    const char *dir;
 {
     struct file_info finfo;
+    char *tmp;
 
     memset (&finfo, 0, sizeof (finfo));
     finfo.file = file;
     finfo.update_dir = dir;
     if (dir[0] == '\0')
-	finfo.fullname = xstrdup (file);
+	tmp = xstrdup (file);
     else
     {
-	finfo.fullname = xmalloc (strlen (file) + strlen (dir) + 10);
-	strcpy (finfo.fullname, dir);
-	strcat (finfo.fullname, "/");
-	strcat (finfo.fullname, file);
+	tmp = xmalloc (strlen (file) + strlen (dir) + 10);
+	strcpy (tmp, dir);
+	strcat (tmp, "/");
+	strcat (tmp, file);
     }
 
+    finfo.fullname = tmp;
     write_letter (&finfo, '?');
-    free (finfo.fullname);
+    free (tmp);
 }
 
+
+
 /* ARGSUSED */
 static int
 update_filesdone_proc (callerdat, err, repository, update_dir, entries)
     void *callerdat;
     int err;
-    char *repository;
-    char *update_dir;
+    const char *repository;
+    const char *update_dir;
     List *entries;
 {
     if (rewrite_tag)
@@ -857,7 +839,7 @@ update_filesdone_proc (callerdat, err, repository, update_dir, entries)
     }
 
     /* Clean up CVS admin dirs if we are export */
-    if (strcmp (command_name, "export") == 0)
+    if (strcmp (cvs_cmd_name, "export") == 0)
     {
 	/* I'm not sure the existence_error is actually possible (except
 	   in cases where we really should print a message), but since
@@ -876,9 +858,11 @@ update_filesdone_proc (callerdat, err, repository, update_dir, entries)
 	    Create_Root ((char *) NULL, current_parsed_root->original);
     }
 
-    return (err);
+    return err;
 }
 
+
+
 /*
  * update_dirent_proc () is called back by the recursion processor before a
  * sub-directory is processed for update.  In this case, update_dirent proc
@@ -890,9 +874,9 @@ update_filesdone_proc (callerdat, err, repository, update_dir, entries)
 static Dtype
 update_dirent_proc (callerdat, dir, repository, update_dir, entries)
     void *callerdat;
-    char *dir;
-    char *repository;
-    char *update_dir;
+    const char *dir;
+    const char *repository;
+    const char *update_dir;
     List *entries;
 {
     if (ignore_directory (update_dir))
@@ -907,7 +891,7 @@ update_dirent_proc (callerdat, dir, repository, update_dir, entries)
     {
 	/* if we aren't building dirs, blow it off */
 	if (!update_build_dirs)
-	    return (R_SKIP_ALL);
+	    return R_SKIP_ALL;
 
 	/* Various CVS administrators are in the habit of removing
 	   the repository directory for things they don't want any
@@ -935,7 +919,7 @@ update_dirent_proc (callerdat, dir, repository, update_dir, entries)
 	{
 	    /* ignore the missing dir if -n is specified */
 	    error (0, 0, "New directory `%s' -- ignored", update_dir);
-	    return (R_SKIP_ALL);
+	    return R_SKIP_ALL;
 	}
 	else
 	{
@@ -1033,9 +1017,11 @@ update_dirent_proc (callerdat, dir, repository, update_dir, entries)
     if (!quiet)
 	error (0, 0, "Updating %s", update_dir);
 
-    return (R_PROCESS);
+    return R_PROCESS;
 }
 
+
+
 /*
  * update_dirleave_proc () is called back by the recursion code upon leaving
  * a directory.  It will prune empty directories if needed and will execute
@@ -1045,13 +1031,11 @@ update_dirent_proc (callerdat, dir, repository, update_dir, entries)
 static int
 update_dirleave_proc (callerdat, dir, err, update_dir, entries)
     void *callerdat;
-    char *dir;
+    const char *dir;
     int err;
-    char *update_dir;
+    const char *update_dir;
     List *entries;
 {
-    FILE *fp;
-
     /* Delete the ignore list if it hasn't already been done.  */
     if (ignlist)
 	dellist (&ignlist);
@@ -1077,45 +1061,6 @@ update_dirleave_proc (callerdat, dir, err, update_dir, entries)
 	tag_update_dir = NULL;
     }
 
-    /* run the update_prog if there is one */
-    /* FIXME: should be checking for errors from CVS_FOPEN and printing
-       them if not existence_error.  */
-    if (err == 0 && !pipeout && !noexec &&
-	(fp = CVS_FOPEN (CVSADM_UPROG, "r")) != NULL)
-    {
-	char *cp;
-	char *repository;
-	char *line = NULL;
-	size_t line_allocated = 0;
-
-	repository = Name_Repository ((char *) NULL, update_dir);
-	if (getline (&line, &line_allocated, fp) >= 0)
-	{
-	    if ((cp = strrchr (line, '\n')) != NULL)
-		*cp = '\0';
-	    run_setup (line);
-	    run_arg (repository);
-	    cvs_output (program_name, 0);
-	    cvs_output (" ", 1);
-	    cvs_output (command_name, 0);
-	    cvs_output (": Executing '", 0);
-	    run_print (stdout);
-	    cvs_output ("'\n", 0);
-	    cvs_flushout ();
-	    (void) run_exec (RUN_TTY, RUN_TTY, RUN_TTY, RUN_NORMAL);
-	}
-	else if (ferror (fp))
-	    error (0, errno, "cannot read %s", CVSADM_UPROG);
-	else
-	    error (0, 0, "unexpected end of file on %s", CVSADM_UPROG);
-
-	if (fclose (fp) < 0)
-	    error (0, errno, "cannot close %s", CVSADM_UPROG);
-	if (line != NULL)
-	    free (line);
-	free (repository);
-    }
-
     if (strchr (dir, '/') == NULL)
     {
 	/* FIXME: chdir ("..") loses with symlinks.  */
@@ -1132,9 +1077,11 @@ update_dirleave_proc (callerdat, dir, err, update_dir, entries)
 	}
     }
 
-    return (err);
+    return err;
 }
 
+
+
 static int isremoved PROTO ((Node *, void *));
 
 /* Returns 1 if the file indicated by node has been removed.  */
@@ -1143,19 +1090,21 @@ isremoved (node, closure)
     Node *node;
     void *closure;
 {
-    Entnode *entdata = (Entnode*) node->data;
+    Entnode *entdata = node->data;
 
     /* If the first character of the version is a '-', the file has been
        removed. */
     return (entdata->version && entdata->version[0] == '-') ? 1 : 0;
 }
 
+
+
 /* Returns 1 if the argument directory is completely empty, other than the
    existence of the CVS directory entry.  Zero otherwise.  If MIGHT_NOT_EXIST
    and the directory doesn't exist, then just return 0.  */
 int
 isemptydir (dir, might_not_exist)
-    char *dir;
+    const char *dir;
     int might_not_exist;
 {
     DIR *dirp;
@@ -1166,7 +1115,7 @@ isemptydir (dir, might_not_exist)
 	if (might_not_exist && existence_error (errno))
 	    return 0;
 	error (0, errno, "cannot open directory %s for empty check", dir);
-	return (0);
+	return 0;
     }
     errno = 0;
     while ((dp = CVS_READDIR (dirp)) != NULL)
@@ -1179,7 +1128,7 @@ isemptydir (dir, might_not_exist)
 		/* An entry other than the CVS directory.  The directory
 		   is certainly not empty. */
 		(void) CVS_CLOSEDIR (dirp);
-		return (0);
+		return 0;
 	    }
 	    else
 	    {
@@ -1210,7 +1159,7 @@ isemptydir (dir, might_not_exist)
 		    /* There are files that have been removed, but not
 		       committed!  Do not consider the directory empty. */
 		    (void) CVS_CLOSEDIR (dirp);
-		    return (0);
+		    return 0;
 		}
 	    }
 	}
@@ -1220,12 +1169,14 @@ isemptydir (dir, might_not_exist)
     {
 	error (0, errno, "cannot read directory %s", dir);
 	(void) CVS_CLOSEDIR (dirp);
-	return (0);
+	return 0;
     }
     (void) CVS_CLOSEDIR (dirp);
-    return (1);
+    return 1;
 }
 
+
+
 /*
  * scratch the Entries file entry associated with a file
  */
@@ -1270,9 +1221,11 @@ scratch_file (finfo, vers)
 	    vers->ts_user = NULL;
 	}
     }
-    return (0);
+    return 0;
 }
 
+
+
 /*
  * Check out a file.
  */
@@ -1359,7 +1312,7 @@ VERS: ", 0);
 	{
 	    revbuf = buf_nonio_initialize ((BUFMEMERRPROC) NULL);
 	    status = RCS_checkout (vers_ts->srcfile, (char *) NULL,
-				   vers_ts->vn_rcs, vers_ts->vn_tag,
+				   vers_ts->vn_rcs, vers_ts->tag,
 				   vers_ts->options, RUN_TTY,
 				   checkout_to_buffer, revbuf);
 	}
@@ -1367,7 +1320,7 @@ VERS: ", 0);
 #endif
 	    status = RCS_checkout (vers_ts->srcfile,
 				   pipeout ? NULL : finfo->file,
-				   vers_ts->vn_rcs, vers_ts->vn_tag,
+				   vers_ts->vn_rcs, vers_ts->tag,
 				   vers_ts->options, RUN_TTY,
 				   (RCSCHECKOUTPROC) NULL, (void *) NULL);
     }
@@ -1390,8 +1343,11 @@ VERS: ", 0);
 		   is here only because noexec doesn't write srcfile->path
 		   for us to stat.  */
 		if (stat (vers_ts->srcfile->path, &sb) < 0)
+		{
+		    buf_free (revbuf);
 		    error (1, errno, "cannot stat %s",
 			   vers_ts->srcfile->path);
+		}
 		mode = sb.st_mode &~ (S_IWRITE | S_IWGRP | S_IWOTH);
 	    }
 
@@ -1497,6 +1453,8 @@ VERS: ", 0);
 	    /* fix up the vers structure, in case it is used by join */
 	    if (join_rev1)
 	    {
+		/* FIXME: Throwing away the original revision info is almost
+		   certainly wrong -- what if join_rev1 is "BASE"?  */
 		if (vers_ts->vn_user != NULL)
 		    free (vers_ts->vn_user);
 		if (vers_ts->vn_rcs != NULL)
@@ -1506,7 +1464,7 @@ VERS: ", 0);
 	    }
 
 	    /* If this is really Update and not Checkout, recode history */
-	    if (strcmp (command_name, "update") == 0)
+	    if (strcmp (cvs_cmd_name, "update") == 0)
 		history_write ('U', finfo->update_dir, xvers_ts->vn_rcs, finfo->file,
 			       finfo->repository);
 
@@ -1553,9 +1511,13 @@ VERS: ", 0);
 	free (backup);
     }
 
-    return (retval);
+    if (revbuf != NULL)
+	buf_free (revbuf);
+    return retval;
 }
 
+
+
 #ifdef SERVER_SUPPORT
 
 /* This function is used to write data from a file being checked out
@@ -1699,8 +1661,20 @@ patch_file (finfo, vers_ts, docheckout, file_info, checksum)
     data.final_nl = 0;
     data.compute_checksum = 0;
 
+    /* FIXME - Passing vers_ts->tag here is wrong in the least number
+     * of cases.  Since we don't know whether vn_user was checked out
+     * using a tag, we pass vers_ts->tag, which, assuming the user did
+     * not specify a new TAG to -r, will be the branch we are on.
+     *
+     * The only thing it is used for is to substitute in for the Name
+     * RCS keyword, so in the error case, the patch fails to apply on
+     * the client end and we end up resending the whole file.
+     *
+     * At least, if we are keeping track of the tag vn_user came from,
+     * I don't know where yet. -DRP
+     */
     retcode = RCS_checkout (vers_ts->srcfile, (char *) NULL,
-			    vers_ts->vn_user, (char *) NULL,
+			    vers_ts->vn_user, vers_ts->tag,
 			    vers_ts->options, RUN_TTY,
 			    patch_file_write, (void *) &data);
 
@@ -1723,7 +1697,7 @@ patch_file (finfo, vers_ts, docheckout, file_info, checksum)
 	cvs_MD5Init (&data.context);
 
 	retcode = RCS_checkout (vers_ts->srcfile, (char *) NULL,
-				vers_ts->vn_rcs, vers_ts->vn_tag,
+				vers_ts->vn_rcs, vers_ts->tag,
 				vers_ts->options, RUN_TTY,
 				patch_file_write, (void *) &data);
 
@@ -1777,46 +1751,60 @@ patch_file (finfo, vers_ts, docheckout, file_info, checksum)
 	{
 	    fail = 1;
 	}
-	else
-	{
+    }
+
+    if (! fail)
+    {
+	struct stat file2_info;
+
+	/* Check to make sure the patch is really shorter */
+	if (CVS_STAT (file2, &file2_info) < 0)
+	    error (1, errno, "could not stat %s", file2);
+	if (CVS_STAT (finfo->file, file_info) < 0)
+	    error (1, errno, "could not stat %s", finfo->file);
+	if (file2_info.st_size <= file_info->st_size)
+	    fail = 1;
+    }
+
+    if (! fail)
+    {
 # define BINARY "Binary"
-	    char buf[sizeof BINARY];
-	    unsigned int c;
+	char buf[sizeof BINARY];
+	unsigned int c;
 
-	    /* Stat the original RCS file, and then adjust it the way
-	       that RCS_checkout would.  FIXME: This is an abstraction
-	       violation.  */
-	    if (CVS_STAT (vers_ts->srcfile->path, file_info) < 0)
-		error (1, errno, "could not stat %s", vers_ts->srcfile->path);
-	    if (chmod (finfo->file,
-		       file_info->st_mode & ~(S_IWRITE | S_IWGRP | S_IWOTH))
-		< 0)
-		error (0, errno, "cannot change mode of file %s", finfo->file);
-	    if (cvswrite
-		&& !fileattr_get (finfo->file, "_watched"))
-		xchmod (finfo->file, 1);
-
-	    /* Check the diff output to make sure patch will be handle it.  */
-	    e = CVS_FOPEN (finfo->file, "r");
-	    if (e == NULL)
-		error (1, errno, "could not open diff output file %s",
-		       finfo->fullname);
-	    c = fread (buf, 1, sizeof BINARY - 1, e);
-	    buf[c] = '\0';
-	    if (strcmp (buf, BINARY) == 0)
-	    {
-		/* These are binary files.  We could use diff -a, but
-		   patch can't handle that.  */
-		fail = 1;
-	    }
-	    fclose (e);
+	/* Check the diff output to make sure patch will be handle it.  */
+	e = CVS_FOPEN (finfo->file, "r");
+	if (e == NULL)
+	    error (1, errno, "could not open diff output file %s",
+		   finfo->fullname);
+	c = fread (buf, 1, sizeof BINARY - 1, e);
+	buf[c] = '\0';
+	if (strcmp (buf, BINARY) == 0)
+	{
+	    /* These are binary files.  We could use diff -a, but
+	       patch can't handle that.  */
+	    fail = 1;
 	}
+	fclose (e);
     }
 
     if (! fail)
     {
         Vers_TS *xvers_ts;
 
+	/* Stat the original RCS file, and then adjust it the way
+	   that RCS_checkout would.  FIXME: This is an abstraction
+	   violation.  */
+	if (CVS_STAT (vers_ts->srcfile->path, file_info) < 0)
+	    error (1, errno, "could not stat %s", vers_ts->srcfile->path);
+	if (chmod (finfo->file,
+		   file_info->st_mode & ~(S_IWRITE | S_IWGRP | S_IWOTH))
+	    < 0)
+	    error (0, errno, "cannot change mode of file %s", finfo->file);
+	if (cvswrite
+	    && !fileattr_get (finfo->file, "_watched"))
+	    xchmod (finfo->file, 1);
+
         /* This stuff is just copied blindly from checkout_file.  I
 	   don't really know what it does.  */
         xvers_ts = Version_TS (finfo, options, tag, date,
@@ -1831,10 +1819,10 @@ patch_file (finfo, vers_ts, docheckout, file_info, checksum)
 	if (CVS_STAT (finfo->file, file_info) < 0)
 	    error (1, errno, "could not stat %s", finfo->file);
 
-	/* If this is really Update and not Checkout, recode history */
-	if (strcmp (command_name, "update") == 0)
-	    history_write ('P', finfo->update_dir, xvers_ts->vn_rcs, finfo->file,
-			   finfo->repository);
+	/* If this is really Update and not Checkout, record history.  */
+	if (strcmp (cvs_cmd_name, "update") == 0)
+	    history_write ('P', finfo->update_dir, xvers_ts->vn_rcs,
+	                   finfo->file, finfo->repository);
 
 	freevers_ts (&xvers_ts);
 
@@ -1871,9 +1859,11 @@ patch_file (finfo, vers_ts, docheckout, file_info, checksum)
     free (backup);
     free (file1);
     free (file2);
-    return (retval);
+    return retval;
 }
 
+
+
 /* Write data to a file.  Record whether the last byte written was a
    newline.  Optionally compute a checksum.  This is called by
    patch_file via RCS_checkout.  */
@@ -1901,7 +1891,7 @@ patch_file_write (callerdat, buffer, len)
  * Several of the types we process only print a bit of information consisting
  * of a single letter and the name.
  */
-static void
+void
 write_letter (finfo, letter)
     struct file_info *finfo;
     int letter;
@@ -1942,6 +1932,8 @@ write_letter (finfo, letter)
     return;
 }
 
+
+
 /*
  * Do all the magic associated with a file which needs to be merged
  */
@@ -2016,8 +2008,8 @@ merge_file (finfo, vers)
 	goto out;
     }
 
-    status = RCS_merge(finfo->rcs, vers->srcfile->path, finfo->file,
-		       vers->options, vers->vn_user, vers->vn_rcs);
+    status = RCS_merge (finfo->rcs, vers->srcfile->path, finfo->file,
+		        vers->options, vers->vn_user, vers->vn_rcs);
     if (status != 0 && status != 1)
     {
 	error (0, status == -1 ? errno : 0,
@@ -2053,6 +2045,8 @@ merge_file (finfo, vers)
     /* fix up the vers structure, in case it is used by join */
     if (join_rev1)
     {
+	/* FIXME: Throwing away the original revision info is almost
+	   certainly wrong -- what if join_rev1 is "BASE"?  */
 	if (vers->vn_user != NULL)
 	    free (vers->vn_user);
 	vers->vn_user = xstrdup (vers->vn_rcs);
@@ -2095,7 +2089,8 @@ merge_file (finfo, vers)
 
 	write_letter (finfo, 'C');
 
-	history_write ('C', finfo->update_dir, vers->vn_rcs, finfo->file, finfo->repository);
+	history_write ('C', finfo->update_dir, vers->vn_rcs, finfo->file,
+	               finfo->repository);
 
     }
     else if (retcode == -1)
@@ -2115,9 +2110,23 @@ merge_file (finfo, vers)
     return retval;
 }
 
+
+
 /*
  * Do all the magic associated with a file which needs to be joined
- * (-j option)
+ * (reached via the -j option to checkout or update).
+ *
+ * INPUTS
+ *   finfo		File information about the destination file.
+ *   vers		The Vers_TS structure for finfo.
+ *
+ * GLOBALS
+ *   join_rev1		From the command line.
+ *   join_rev2		From the command line.
+ *   server_active	Natch.
+ *
+ * ASSUMPTIONS
+ *   1.  Is not called in client mode.
  */
 static void
 join_file (finfo, vers)
@@ -2168,6 +2177,9 @@ join_file (finfo, vers)
 	jdate1 = NULL;
     }
 
+    /* FIXME: Need to handle "BASE" for jrev1 and/or jrev2.  Note caveat
+       below about vn_user.  */
+
     /* Convert the second revision, walking branches and dates.  */
     rev2 = RCS_getversion (vers->srcfile, jrev2, jdate2, 1, (int *) NULL);
 
@@ -2349,13 +2361,41 @@ join_file (finfo, vers)
 	return;
     }
 
-    /* If the target of the merge is the same as the working file
-       revision, then there is nothing to do.  */
-    if (vers->vn_user != NULL && strcmp (rev2, vers->vn_user) == 0)
+    /* If the two merge revisions are the same, then there is nothing
+     * to do.  This needs to be checked before the rev2 == up-to-date base
+     * revision check tha comes next.  Otherwise, rev1 can == rev2 and get an
+     * "already contains the changes between <rev1> and <rev1>" message.
+     */
+    if (rev1 && strcmp (rev1, rev2) == 0)
     {
+	free (rev1);
+	free (rev2);
+	return;
+    }
+
+    /* If we know that the user file is up-to-date, then it becomes an
+     * optimization to skip the merge when rev2 is the same as the base
+     * revision.  i.e. we know that diff3(file2,file1,file2) will produce
+     * file2.
+     */
+    if (vers->vn_user != NULL && vers->ts_user != NULL
+        && strcmp (vers->ts_user, vers->ts_rcs) == 0
+        && strcmp (rev2, vers->vn_user) == 0)
+    {
+	if (!really_quiet)
+	{
+	    cvs_output (finfo->fullname, 0);
+	    cvs_output (" already contains the differences between ", 0);
+	    cvs_output (rev1 ? rev1 : "creation", 0);
+	    cvs_output (" and ", 0);
+	    cvs_output (rev2, 0);
+	    cvs_output ("\n", 1);
+	}
+
 	if (rev1 != NULL)
 	    free (rev1);
 	free (rev2);
+
 	return;
     }
 
@@ -2410,17 +2450,8 @@ join_file (finfo, vers)
 	return;
     }
 
-    /* If the two merge revisions are the same, then there is nothing
-       to do.  */
-    if (strcmp (rev1, rev2) == 0)
-    {
-	free (rev1);
-	free (rev2);
-	return;
-    }
-
     /* If there is no working file, then we can't do the merge.  */
-    if (vers->vn_user == NULL)
+    if (vers->vn_user == NULL || vers->vn_user[0] == '-')
     {
 	free (rev1);
 	free (rev2);
@@ -2444,8 +2475,11 @@ join_file (finfo, vers)
     {
 	int retcode;
 	/* The file is up to date.  Need to check out the current contents.  */
+	/* FIXME - see the FIXME comment above the call to RCS_checkout in the
+	 * patch_file function.
+	 */
 	retcode = RCS_checkout (vers->srcfile, finfo->file,
-				vers->vn_user, (char *) NULL,
+				vers->vn_user, vers->tag,
 				(char *) NULL, RUN_TTY,
 				(RCSCHECKOUTPROC) NULL, (void *) NULL);
 	if (retcode != 0)
@@ -2491,16 +2525,29 @@ join_file (finfo, vers)
 	&& vers->ts_user != NULL
 	&& strcmp (vers->ts_user, vers->ts_rcs) == 0
 
-	/* This is because of the worry below about $Name.  If that
-	   isn't a problem, I suspect this code probably works for
-	   text files too.  */
+	/* Avoid this in the text file case.  See below for why.
+	 */
 	&& (strcmp (t_options, "-kb") == 0
 	    || wrap_merge_is_copy (finfo->file)))
     {
-	/* FIXME: what about nametag?  What does RCS_merge do with
-	   $Name?  */
-	if (RCS_checkout (finfo->rcs, finfo->file, rev2, NULL, t_options,
-			  RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0)
+	/* FIXME: Verify my comment below:
+	 *
+	 * RCS_merge does nothing with keywords.  It merges the changes between
+	 * two revisions without expanding the keywords (it might expand in
+	 * -kk mode before computing the diff between rev1 and rev2 - I'm not
+	 * sure).  In other words, the keyword lines in the current work file
+	 * get left alone.
+	 *
+	 * Therfore, checking out the destination revision (rev2) is probably
+	 * incorrect in the text case since we should see the keywords that were
+	 * substituted into the original file at the time it was checked out
+	 * and not the keywords from rev2.
+	 *
+	 * Also, it is safe to pass in NULL for nametag since we know no
+	 * substitution is happening during the binary mode checkout.
+	 */
+	if (RCS_checkout ( finfo->rcs, finfo->file, rev2, (char *)NULL, t_options,
+			   RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0 )
 	    status = 2;
 	else
 	    status = 0;
@@ -2528,13 +2575,14 @@ join_file (finfo, vers)
 	     || special_file_mismatch (finfo, rev1, rev2))
     {
 	/* We are dealing with binary files, or files with a
-	   permission/linkage mismatch, and real merging would
+	   permission/linkage mismatch (this second case only occurs when
+	   PRESERVE_PERMISSIONS_SUPPORT is enabled), and real merging would
 	   need to take place.  This is a conflict.  We give the user
 	   the two files, and let them resolve it.  It is possible
 	   that we should require a "touch foo" or similar step before
 	   we allow a checkin.  */
-	if (RCS_checkout (finfo->rcs, finfo->file, rev2, NULL, t_options,
-			  RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0)
+	if (RCS_checkout ( finfo->rcs, finfo->file, rev2, (char *)NULL,
+			   t_options, RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0)
 	    status = 2;
 	else
 	    status = 0;
@@ -2566,16 +2614,39 @@ join_file (finfo, vers)
 	status = RCS_merge (finfo->rcs, vers->srcfile->path, finfo->file,
 			    t_options, rev1, rev2);
 
-    if (status != 0 && status != 1)
+    if (status != 0)
     {
-	error (0, status == -1 ? errno : 0,
-	       "could not merge revision %s of %s", rev2, finfo->fullname);
-	error (status == -1 ? 1 : 0, 0, "restoring %s from backup file %s",
-	       finfo->fullname, backup);
-	rename_file (backup, finfo->file);
+	if (status != 1)
+	{
+	    error (0, status == -1 ? errno : 0,
+		   "could not merge revision %s of %s", rev2, finfo->fullname);
+	    error (status == -1 ? 1 : 0, 0, "restoring %s from backup file %s",
+		   finfo->fullname, backup);
+	    rename_file (backup, finfo->file);
+	}
+    }
+    else /* status == 0 */
+    {
+	/* FIXME: the noexec case is broken.  RCS_merge could be doing the
+	   xcmp on the temporary files without much hassle, I think.  */
+	if (!noexec && !xcmp (backup, finfo->file))
+	{
+	    if (!really_quiet)
+	    {
+		cvs_output (finfo->fullname, 0);
+		cvs_output (" already contains the differences between ", 0);
+		cvs_output (rev1, 0);
+		cvs_output (" and ", 0);
+		cvs_output (rev2, 0);
+		cvs_output ("\n", 1);
+	    }
+
+	    /* and skip the registering and sending the new file since it
+	     * hasn't been updated.
+	     */
+	    goto out;
+	}
     }
-    free (rev1);
-    free (rev2);
 
     /* The file has changed, but if we just checked it out it may
        still have the same timestamp it did when it was first
@@ -2612,9 +2683,15 @@ join_file (finfo, vers)
 			(struct buffer *) NULL);
     }
 #endif
+
+out:
+    free (rev1);
+    free (rev2);
     free (backup);
 }
 
+
+
 /*
  * Report whether revisions REV1 and REV2 of FINFO agree on:
  *   . file ownership
@@ -2675,7 +2752,7 @@ special_file_mismatch (finfo, rev1, rev2)
 	    rev1_symlink = xreadlink (finfo->file);
 	else
 	{
-# ifdef HAVE_ST_RDEV
+# ifdef HAVE_STRUCT_STAT_ST_RDEV
 	    if (CVS_LSTAT (finfo->file, &sb) < 0)
 		error (1, errno, "could not get file information for %s",
 		       finfo->file);
@@ -2694,7 +2771,7 @@ special_file_mismatch (finfo, rev1, rev2)
     else
     {
 	n = findnode (finfo->rcs->versions, rev1);
-	vp = (RCSVers *) n->data;
+	vp = n->data;
 
 	n = findnode (vp->other_delta, "symlink");
 	if (n != NULL)
@@ -2729,7 +2806,7 @@ special_file_mismatch (finfo, rev1, rev2)
 		if (sscanf (n->data, "%15s %lu", ftype,
 			    &dev_long) < 2)
 		    error (1, 0, "%s:%s has bad `special' newphrase %s",
-			   finfo->file, rev1, n->data);
+			   finfo->file, rev1, (char *)n->data);
 		rev1_dev = dev_long;
 		if (strcmp (ftype, "character") == 0)
 		    rev1_mode |= S_IFCHR;
@@ -2753,7 +2830,7 @@ special_file_mismatch (finfo, rev1, rev2)
 	    rev2_symlink = xreadlink (finfo->file);
 	else
 	{
-# ifdef HAVE_ST_RDEV
+# ifdef HAVE_STRUCT_STAT_ST_RDEV
 	    if (CVS_LSTAT (finfo->file, &sb) < 0)
 		error (1, errno, "could not get file information for %s",
 		       finfo->file);
@@ -2772,7 +2849,7 @@ special_file_mismatch (finfo, rev1, rev2)
     else
     {
 	n = findnode (finfo->rcs->versions, rev2);
-	vp = (RCSVers *) n->data;
+	vp = n->data;
 
 	n = findnode (vp->other_delta, "symlink");
 	if (n != NULL)
@@ -2807,7 +2884,7 @@ special_file_mismatch (finfo, rev1, rev2)
 		if (sscanf (n->data, "%15s %lu", ftype,
 			    &dev_long) < 2)
 		    error (1, 0, "%s:%s has bad `special' newphrase %s",
-			   finfo->file, rev2, n->data);
+			   finfo->file, rev2, (char *)n->data);
 		rev2_dev = dev_long;
 		if (strcmp (ftype, "character") == 0)
 		    rev2_mode |= S_IFCHR;
@@ -2926,8 +3003,10 @@ special_file_mismatch (finfo, rev1, rev2)
 #endif
 }
 
+
+
 int
 joining ()
 {
-    return (join_rev1 != NULL);
+    return join_rev1 != NULL;
 }
